#include <stdio.h>
#include <sys/socket.h>
#include <sys/types.h>
#include <arpa/inet.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <netinet/in.h>
#include <pthread.h>
#include <limits.h>
#include <linux/netfilter_ipv4.h>
#include <errno.h>
#include <time.h>

#define PORT 34557
#define MAXPENDING 15

#define CLIENT_IN_ADDR inet_addr("0.0.0.0")
#define OUT_ADDR INADDR_ANY

#define MAX_MSG_SIZE 16384

#define BUF_SIZE 1600000

struct thread_params {
  struct sockaddr_in client_in;
  unsigned int clientlen;
  pthread_t t_proxyClient;
  int clientfd;
  int remotefd;
};

struct bytes_entry {
  char bytes[MAX_MSG_SIZE];
  unsigned int len;
};

void Die(const char *mess) { 
  perror(mess); 
  exit(1); 
}

void DieThread(const char *mess) { 
  perror(mess);
  pthread_exit((void *)NULL); 
}

static void heandleClient_cleanup(void *data) {
  struct thread_params *params = (struct thread_params *)data;
  close(params->clientfd);
  
  if (params->remotefd > 0) {
    close(params->remotefd);
  }
  
  free(params);
}

void forward_bytes(int readfd, int writefd, short *read_closed, short *write_closed) {
  struct bytes_entry bytes;
  bytes.len = 0;
  bytes.len = recv(readfd, bytes.bytes, sizeof bytes.bytes, 0);

  if ((bytes.len <= 0 && errno != EINTR) || bytes.len > sizeof bytes.bytes) {
    #ifdef DEBUG
    printf("Closing read \n");
    #endif
    *read_closed = 1;
    shutdown(readfd, SHUT_RD);
  }
  #ifdef DEBUG
  printf("Read %u bytes \n", bytes.len);
  #endif
  
  if ((bytes.len <= 0 && errno != EINTR) || bytes.len > sizeof bytes.bytes) {
    #ifdef DEBUG
    printf("Closing write \n");
    #endif
    *write_closed = 1;
    shutdown(writefd, SHUT_WR);
  }
  
  if (*write_closed) {
    return;
  }
  
  int result = send(writefd, bytes.bytes, bytes.len, 0);
  
  #ifdef DEBUG
  printf("Sent %u bytes \n", result);
  #endif
  
  if (result != bytes.len) {
    #ifdef DEBUG
    printf("Closing write \n");
    #endif
    *write_closed = 1;
    shutdown(writefd, SHUT_WR);
  }
}

int get_remotefd(unsigned int dst_ip, unsigned short dst_port, struct sockaddr_in *client_addr) {
  int remotefd = -1;
  struct sockaddr_in remote_in, out_in;
  
  if ((remotefd = socket(PF_INET, SOCK_STREAM, IPPROTO_TCP)) < 0) {
    printf("Error creating socket\n");
    return remotefd;
  }
  
  memset(&remote_in, 0, sizeof(remote_in));
  remote_in.sin_family = AF_INET;
  remote_in.sin_port = htons(dst_port);
  remote_in.sin_addr.s_addr = htonl(dst_ip);
  
  memset(&out_in, 0, sizeof(out_in));
  out_in.sin_family = AF_INET;
  out_in.sin_addr.s_addr = OUT_ADDR;

  #ifdef DEBUG
  printf("Attempting to connect to %s:%u\n", inet_ntoa(remote_in.sin_addr), dst_port);
  #endif

  // Bind first here
  if (bind(remotefd, (struct sockaddr *) &out_in, sizeof(struct sockaddr_in)) < 0) {
    printf("Failed to bind the remote socket\n");
    close(remotefd);
    return -1;
  }
  
  // Establish connection
  if (connect(remotefd, (struct sockaddr *) &remote_in, sizeof(struct sockaddr_in)) < 0) {
    printf("Failed to connect with server\n");
    close(remotefd);
    return -1;
  }
  
  #ifdef DEBUG
  printf("Connection succeeded\n");
  #endif

  return remotefd;
}

void do_proxy(int clientfd, int remotefd) {
  int maxfd, rv;
  fd_set rsock_fds;
  short client_rclosed = 0;
  short client_wclosed = 0;
  short remote_rclosed = 0;
  short remote_wclosed = 0;
    
  maxfd = clientfd > remotefd ? clientfd : remotefd;

  while(1) {
    #ifdef DEBUG
    printf("client_rclosed %u client_wclosed %u remote_rclosed %u remote_wclosed %u \n", 
	  client_rclosed, client_wclosed, remote_rclosed, remote_wclosed);
    #endif
  
    if (client_rclosed && remote_rclosed) {
      break;
    }
  
    FD_ZERO(&rsock_fds);
	
    if (!client_rclosed) {
      FD_SET(clientfd, &rsock_fds);
    }

    
    if (!remote_rclosed) {
      FD_SET(remotefd, &rsock_fds);
    }

	
    rv = select(maxfd + 1, &rsock_fds, NULL, NULL, NULL);

    if (rv == -1 && errno == EINTR) continue;

    if (rv == -1 && errno == EBADF) {
      break;
    }

    if(FD_ISSET(clientfd, &rsock_fds) && !(client_rclosed || remote_wclosed)) {
      #ifdef DEBUG
      printf("Forwarding bytes from client to remote\n");
      #endif
      forward_bytes(clientfd, remotefd, &client_rclosed, &remote_wclosed);
    }
	
    if(FD_ISSET(remotefd, &rsock_fds) && !(remote_rclosed || client_wclosed)) {
      #ifdef DEBUG
      printf("Forwarding bytes from remote to client\n");
      #endif
      forward_bytes(remotefd, clientfd, &remote_rclosed, &client_wclosed);
    }
  }
    
}

void *handleClient(void *data) {
  struct thread_params *params = (struct thread_params *)data;
  int clientfd = params->clientfd;
  struct sockaddr_in client_in = params->client_in;
  int remotefd, buf_size;
  unsigned short dst_port = 0;
  unsigned int dst_ip = 0;
  
  struct sockaddr_in orig_addr_in;
  socklen_t addrlen = sizeof(orig_addr_in);

  buf_size = BUF_SIZE;

  pthread_cleanup_push(heandleClient_cleanup, data);
  
  memset(&orig_addr_in, 0, sizeof(orig_addr_in));
  
  if(getsockopt(clientfd, IPPROTO_IP, SO_ORIGINAL_DST, &orig_addr_in, &addrlen)) {
    if(getsockname(clientfd, (struct sockaddr *) &orig_addr_in, &addrlen)) {
      DieThread("Error getting original client destination");
    }
  }
  

  dst_port = ntohs(orig_addr_in.sin_port);
  #ifdef DEBUG
  printf("Received dst_port %u \n", dst_port);
  #endif

  dst_ip = ntohl(orig_addr_in.sin_addr.s_addr);
  #ifdef DEBUG
  printf("Received dst_ip %u \n", dst_ip);
  #endif
  
  if (dst_ip == 0) {
    DieThread("Failed to receive original client destination");
  }
  
  remotefd = get_remotefd(dst_ip, dst_port, &client_in);

  if (remotefd <= 0) {
    DieThread("Error connecting to or resolving host");
  }
  
  params->remotefd = remotefd; //for cleanup


  do_proxy(clientfd, remotefd);

  pthread_cleanup_pop(0);
  heandleClient_cleanup(data);
  return NULL;
}

int main(int argc, char *argv[]) {
  int serverfd, buf_size;
  int value = 1;
  struct sockaddr_in server_in;

  buf_size = BUF_SIZE;

  if ((serverfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) {
    Die("Failed to create socket");
  }

  memset(&server_in, 0, sizeof(server_in));

  server_in.sin_family = AF_INET;
  server_in.sin_addr.s_addr = CLIENT_IN_ADDR;
  server_in.sin_port = htons(PORT);


  if (setsockopt(serverfd, SOL_SOCKET, SO_REUSEADDR, (char *) &value, sizeof(value)) < 0) {
    Die("Could not set max socket buffer size");
  }

  if (bind(serverfd, (struct sockaddr *) &server_in, sizeof(server_in)) < 0) {
    Die("Failed to bind the server socket");
  }
  

  if (listen(serverfd, MAXPENDING) < 0) {
    Die("Failed to listen on server socket");
  }

  while (1) {
    struct thread_params *params = (struct thread_params *) malloc(sizeof(struct thread_params));
    memset(params, 0, sizeof(struct thread_params));
    params->clientlen = sizeof(struct sockaddr_in);
	
    if ((params->clientfd = accept(serverfd, (struct sockaddr *) &(params->client_in), &(params->clientlen))) < 0) {
      Die("Failed to accept client connection");
    }
    
    #ifdef DEBUG
    printf("Client connected: %s\n", inet_ntoa(params->client_in.sin_addr));
    #endif
    
    pthread_create( &(params->t_proxyClient), NULL, handleClient, (void*) params);
    
  }
  
  return 0;
}
