/* -*- Mode:C++; c-file-style:"gnu"; indent-tabs-mode:nil; -*- */
/*
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License version 2 as
 * published by the Free Software Foundation;
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * Copyright 2011 Rob Smits, Divam Jain, Sarah Pidcock, Ian Goldberg
 */

// Standard C++ libraries
#include <iostream>
#include <iomanip>
#include <sstream>
#include <map>

// C TCP/IP libraries
#include <netinet/in.h>
#include <netinet/tcp.h>   //Provides declarations for tcp header
#include <netinet/ip.h>    //Provides declarations for ip header
#include <netdb.h>
#include <arpa/inet.h>

// Misc C libraries
#include <time.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <pthread.h>

#include <linux/netfilter.h>  /* Defines verdicts (NF_ACCEPT, etc) */
#include <libnetfilter_queue/libnetfilter_queue.h>

//BridgeSPA includes
#include "crypto.c"
#include "utils.cpp"

using namespace std;

time_t last_calculated_mackey = 0;
time_t curr_minute;
uint32_t update_frequency = 0;
unsigned char curr_mackey[32];
unsigned char* seed_key;
int seed_key_len = 0;

pthread_mutex_t mackey_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t minutes_mutex = PTHREAD_MUTEX_INITIALIZER;

#define MAX_MSG_SIZE 16384

void copyMACKey(unsigned char bridgekey[32]) {
  //assume bridgekey has already been zeroed
  int looplimit = 32;

  for (int i = 0; i < looplimit; i++) {
    bridgekey[i] = curr_mackey[i];
  }
}

// If we enforce the length of the seed_key we can remove this basically 
// duplicate function

void copySeedKey(unsigned char bridgekey[32]) {
  //assume bridgekey has already been zeroed
  int looplimit = seed_key_len <= 32 ? seed_key_len : 32;

  for (int i = 0; i < looplimit; i++) {
    bridgekey[i] = seed_key[i];
  }
} 

void getCurrentMACKey() {
  SharedSecret bridgeSecret;
  stringstream ss (stringstream::in | stringstream::out);
  string mac_data;
  time_t epoch_index;
  time_t last_calculated_index = last_calculated_mackey / update_frequency;
  
  epoch_index = time(NULL) / update_frequency;

  if (epoch_index == last_calculated_index) {
    #ifdef DEBUG
    cout << "Using previously-calculated MACKey: " << curr_mackey << endl;
    #endif
    return;
  }

  memset(bridgeSecret.key, 0, sizeof(bridgeSecret.key));
  copySeedKey(bridgeSecret.key);

  ss << epoch_index;
  mac_data = ss.str();

  generateMACKey(bridgeSecret, (unsigned char*)mac_data.c_str(), mac_data.size(), curr_mackey);

  last_calculated_index = epoch_index;
}

void* maintainMACKey(void *data) {
  while(1) {
    pthread_mutex_lock(&mackey_mutex);
    getCurrentMACKey();
    pthread_mutex_unlock(&mackey_mutex);
    sleep(60);
  }  
}

void* maintainMinutes(void *data) {
  while(1) {
    pthread_mutex_lock(&minutes_mutex);
    curr_minute = (time(NULL)/60);
    pthread_mutex_unlock(&minutes_mutex);
    sleep(20);
  }  
}


uint32_t getTimestamp(struct tcphdr *tcp_hdr, uint16_t &timestamp_found) {
  unsigned char *options_ptr = (unsigned char *)(tcp_hdr) + 20;
  uint16_t data_offset = *((uint8_t*)((unsigned char *)(tcp_hdr) + 12));
  uint32_t *long_ptr;
  uint32_t timestamp = 0;
  timestamp_found = 0;

  data_offset = data_offset >> 4;  //only first 4 bits
  data_offset *= 4;

  cout << hex;

  while(1) {
    uint8_t kind = (uint8_t)(*options_ptr);
    options_ptr++;

    if (kind == 1) {
      continue;
    }

    uint8_t len = (uint8_t)*options_ptr;
    options_ptr++;


    if (kind == 8) {
      unsigned char old_byte = 0;

      long_ptr = (uint32_t*) options_ptr;
      timestamp = ntohl(*long_ptr);
      
      timestamp_found = 1;
      break;

    } else if (kind == 0) {
      break;
    } else {
      options_ptr += len - 2;
    }

    if (((uint64_t)options_ptr) - ((uint64_t)tcp_hdr) >= data_offset) {
      break;
    }
  }

  cout << dec;

  return timestamp;
}

// libnetfilter_queue code based on tutorial from Joe Kopena
static int callback(nfq_q_handle *myQueue, struct nfgenmsg *msg,
                    nfq_data *pkt, void *cbData) {
  uint32_t packet_id = 0;
  nfqnl_msg_packet_hdr *header;
  struct iphdr *ip_hdr;
  struct tcphdr *tcp_hdr;
 
  struct sockaddr_in source_ip, dest_ip;
  uint32_t seq;  
  uint32_t spa_conn_tag = 0;
  uint32_t client_conn_tag = 0;
  uint32_t client_timestamp = 0;

  uint32_t dest_ip_n, src_ip_n;       // Used for
  uint16_t dest_port, src_port;       // debugging

  SharedSecret bridgeSecret;

  time_t minutes = 0;

  uint16_t timestamp_found = 0;

  header = nfq_get_msg_packet_hdr(pkt);
  packet_id = ntohl(header->packet_id);
  
  // Print the payload; in copy meta mode, only headers will be included;
  // in copy packet mode, whole packet will be returned.
  #ifdef PAYLOAD_UCHAR
  unsigned char *pktData;
  #else
  char *pktData;
  #endif
  int len = nfq_get_payload(pkt, &pktData);
  
  if (!len) {
    return nfq_set_verdict(myQueue, packet_id, NF_DROP, 0, NULL);
  }

  memset(&source_ip, 0, sizeof(source_ip));
  memset(&dest_ip, 0, sizeof(dest_ip));

  ip_hdr = (struct iphdr *) pktData;

  tcp_hdr = (struct tcphdr *) (pktData + sizeof(struct iphdr)); 

  source_ip.sin_addr.s_addr = ip_hdr->saddr;
  dest_ip.sin_addr.s_addr = ip_hdr->daddr;

  src_ip_n = ntohl(source_ip.sin_addr.s_addr);
  dest_ip_n = ntohl(dest_ip.sin_addr.s_addr);
  src_port = ntohs(tcp_hdr->source);
  dest_port = ntohs(tcp_hdr->dest);

  seq = ntohl(tcp_hdr->seq);

  #ifdef DEBUG
  cout << "Source Address: " << inet_ntoa(source_ip.sin_addr) << endl;
  cout << "Destination Address: " << inet_ntoa(dest_ip.sin_addr) << endl;
  cout << "Source Port: " << (uint32_t)src_port << endl;
  cout << "Destination Port: " << (uint32_t)dest_port << endl;
  cout << "Sequence Number: " << hex << seq << endl;
  cout << "Acknowledgement Number: " << (unsigned int)ntohl(tcp_hdr->ack_seq) << endl;
  cout << "TCP Checksum: " << (unsigned short)ntohs(tcp_hdr->check) << endl;
  #endif

  client_timestamp = getTimestamp(tcp_hdr, timestamp_found);

  if (timestamp_found) {
    client_conn_tag = (seq << 8) | (client_timestamp & 0x00FF);
  } else {
    client_conn_tag = seq;
  }

  #ifdef DEBUG
  cout << "Client conn tag: " << hex << client_conn_tag << dec << endl;
  #endif

  memset(bridgeSecret.key, 0, sizeof(bridgeSecret.key));

  pthread_mutex_lock(&mackey_mutex);
  copyMACKey(bridgeSecret.key);
  pthread_mutex_unlock(&mackey_mutex);

  //XXX: Yes this is odd -- replicating key alignment behaviour of knockproxy
  string key = hex_binary_to_str(bridgeSecret.key, 32);
  memset(bridgeSecret.key, 0, sizeof(bridgeSecret.key));
  copyKey(key, bridgeSecret.key);

  pthread_mutex_lock(&minutes_mutex);
  minutes = curr_minute;
  pthread_mutex_unlock(&minutes_mutex);

  spa_conn_tag = getConnectionTag(bridgeSecret, source_ip, dest_ip, tcp_hdr, ip_hdr, minutes);

  if (spa_conn_tag == client_conn_tag) {
    #ifdef DEBUG
    cout << "Connection tag matches... accepting" << endl;
    #endif
    return nfq_set_verdict(myQueue, packet_id, NF_ACCEPT, 0, (unsigned char*)NULL);
  } 

  spa_conn_tag = getConnectionTag(bridgeSecret, source_ip, dest_ip, tcp_hdr, ip_hdr, minutes+1);

  
  if (spa_conn_tag == client_conn_tag) {
    #ifdef DEBUG
    cout << "Connection tag matches... accepting" << endl;
    #endif
    return nfq_set_verdict(myQueue, packet_id, NF_ACCEPT, 0, (unsigned char*)NULL);
  } 

  spa_conn_tag = getConnectionTag(bridgeSecret, source_ip, dest_ip, tcp_hdr, ip_hdr, minutes-1);

  
  if (spa_conn_tag == client_conn_tag) {
    #ifdef DEBUG
    cout << "Connection tag matches... accepting" << endl;
    #endif 
    return nfq_set_verdict(myQueue, packet_id, NF_ACCEPT, 0, (unsigned char*)NULL); 
  } 

  #ifdef DEBUG
  cout << "Connection tag didn't match... dropping" << endl;
  cout << endl;
  #endif

  return nfq_set_verdict(myQueue, packet_id, NF_DROP, 0, (unsigned char*)NULL);
}

// libnetfilter_queue code based on tutorial from Joe Kopena
void* processQueue(void *data) {
  struct nfq_handle *nfqHandle;
  struct nfq_q_handle *myQueue;
  struct nfnl_handle *netlinkHandle;

  int fd, res;
  char buf[MAX_MSG_SIZE]; // XXX: is this sufficient?

  // Get a queue connection handle from the module
  if (!(nfqHandle = nfq_open())) {
    cerr << "Error in nfq_open()" << endl;
   return NULL;
  }

  // Unbind the handler from processing any IP packets
  // Not totally sure why this is done, or if it's necessary...
  if (nfq_unbind_pf(nfqHandle, AF_INET) < 0) {
    cerr << "Error in nfq_unbind_pf()" << endl;
  }

  // Bind this handler to process IP packets...
  if (nfq_bind_pf(nfqHandle, AF_INET) < 0) {
    cerr << "Error in nfq_bind_pf()" << endl;
  } // This will often error if queue was previously bound, but will still work

  // Install a callback on queue 0
  if (!(myQueue = nfq_create_queue(nfqHandle,  0, &callback, NULL))) {
    cerr << "Error in nfq_create_queue()" << endl;
    return NULL;
  }

  // Turn on packet copy mode
  if (nfq_set_mode(myQueue, NFQNL_COPY_PACKET, 0xffff) < 0) {
    cerr << "Could not set packet copy mode" << endl;
    return NULL;
  }

  netlinkHandle = nfq_nfnlh(nfqHandle);
  fd = nfnl_fd(netlinkHandle);

  while ((res = recv(fd, buf, sizeof(buf), 0)) && res >= 0) {
    nfq_handle_packet(nfqHandle, buf, res);
  }

  close(fd);
  nfq_destroy_queue(myQueue);

  nfq_close(nfqHandle);
  return 0;
}

int main(int argc, char **argv) {
  if (argc > 2) {
    stringstream ss (stringstream::in | stringstream::out);

    char* str_seed_key = argv[1];
    seed_key_len = strlen(str_seed_key);
    seed_key = (unsigned char*) malloc(sizeof(unsigned char) * seed_key_len);
    hex_to_binary(str_seed_key, seed_key, &seed_key_len);

    ss << argv[2];
    ss >> update_frequency;

  } else {
    cout << "Usage: " << argv[0] << " <SeedKey (hex)> <Update frequency>" << endl;
    exit(0);
  }

  pthread_t t_maintain_mackey, t_maintain_minutes;

  // To avoid system calls in every callback, we maintain the mackey and
  // current minutes value in a different thread
  pthread_create( &t_maintain_mackey, NULL, maintainMACKey, (void*) 0);
  pthread_create( &t_maintain_minutes, NULL, maintainMinutes, (void*) 0);

  processQueue(NULL);
}
